@@ -19,9 +19,7 @@
 
 #include "query.h"
 #include "settings.h"
-
-#include <exo/exo.h>
-#include <libxfce4ui/libxfce4ui.h>
+#include "util.h"
 
 using namespace WhiskerMenu;
 
@@ -54,9 +52,10 @@ static std::string normalize(const gchar* string)
 
 //-----------------------------------------------------------------------------
 
+#if !LIBXFCE4UTIL_CHECK_VERSION(4,15,2)
 static void replace_with_quoted_string(std::string& command, std::string::size_type& index, const gchar* unquoted)
 {
-	if (!exo_str_is_empty(unquoted))
+	if (!xfce_str_is_empty(unquoted))
 	{
 		gchar* quoted = g_shell_quote(unquoted);
 		command.replace(index, 2, quoted);
@@ -73,7 +72,7 @@ static void replace_with_quoted_string(std::string& command, std::string::size_t
 
 static void replace_with_quoted_string(std::string& command, std::string::size_type& index, const gchar* prefix, const gchar* unquoted)
 {
-	if (!exo_str_is_empty(unquoted))
+	if (!xfce_str_is_empty(unquoted))
 	{
 		command.replace(index, 2, prefix);
 		index += strlen(prefix);
@@ -91,11 +90,57 @@ static void replace_with_quoted_string(std::string& command, std::string::size_t
 
 //-----------------------------------------------------------------------------
 
-static void replace_and_free_with_quoted_string(std::string& command, std::string::size_type& index, gchar* unquoted)
+static gchar* xfce_expand_desktop_entry_field_codes(const gchar* command, GSList* /*uri_list*/,
+		const gchar* icon, const gchar* name, const gchar* uri, gboolean requires_terminal)
 {
-	replace_with_quoted_string(command, index, unquoted);
-	g_free(unquoted);
+	std::string expanded(command);
+
+	if (requires_terminal)
+	{
+		expanded.insert(0, "exo-open --launch TerminalEmulator ");
+	}
+
+	std::string::size_type length = expanded.length() - 1;
+	for (std::string::size_type i = 0; i < length; ++i)
+	{
+		if (G_UNLIKELY(command[i] == '%'))
+		{
+			switch (command[i + 1])
+			{
+			case 'i':
+				replace_with_quoted_string(expanded, i, "--icon ", icon);
+				break;
+
+			case 'c':
+				replace_with_quoted_string(expanded, i, name);
+				break;
+
+			case 'k':
+				replace_with_quoted_string(expanded, i, uri);
+				break;
+
+			case '%':
+				expanded.erase(i, 1);
+				break;
+
+			case 'f':
+				// unsupported, pass in a single file dropped on launcher
+			case 'F':
+				// unsupported, pass in a list of files dropped on launcher
+			case 'u':
+				// unsupported, pass in a single URL dropped on launcher
+			case 'U':
+				// unsupported, pass in a list of URLs dropped on launcher
+			default:
+				expanded.erase(i, 2);
+				break;
+			}
+			length = expanded.length() - 1;
+		}
+	}
+	return g_strdup(expanded.c_str());
 }
+#endif
 
 //-----------------------------------------------------------------------------
 
@@ -120,7 +165,7 @@ Launcher::Launcher(GarconMenuItem* item) :
 		generic_name = "";
 	}
 
-	if (!wm_settings->launcher_show_name && !exo_str_is_empty(generic_name))
+	if (!wm_settings->launcher_show_name && !xfce_str_is_empty(generic_name))
 	{
 		std::swap(name, generic_name);
 	}
@@ -155,7 +200,7 @@ Launcher::Launcher(GarconMenuItem* item) :
 	for (GList* i = keywords; i; i = i->next)
 	{
 		const gchar* keyword = static_cast<gchar*>(i->data);
-		if (!exo_str_is_empty(keyword) && g_utf8_validate(keyword, -1, nullptr))
+		if (!xfce_str_is_empty(keyword) && g_utf8_validate(keyword, -1, nullptr))
 		{
 			m_search_keywords.push_back(normalize(keyword));
 		}
@@ -164,7 +209,7 @@ Launcher::Launcher(GarconMenuItem* item) :
 
 	// Create search text for command
 	const gchar* command = garcon_menu_item_get_command(m_item);
-	if (!exo_str_is_empty(command) && g_utf8_validate(command, -1, nullptr))
+	if (!xfce_str_is_empty(command) && g_utf8_validate(command, -1, nullptr))
 	{
 		m_search_command = normalize(command);
 	}
@@ -245,153 +290,58 @@ void Launcher::hide()
 
 void Launcher::run(GdkScreen* screen) const
 {
+	// Expand the field codes
 	const gchar* string = garcon_menu_item_get_command(m_item);
-	if (exo_str_is_empty(string))
+	if (xfce_str_is_empty(string))
 	{
 		return;
 	}
-	std::string command(string);
 
-	if (garcon_menu_item_requires_terminal(m_item))
-	{
-		command.insert(0, "exo-open --launch TerminalEmulator ");
-	}
-
-	// Expand the field codes
-	std::string::size_type length = command.length() - 1;
-	for (std::string::size_type i = 0; i < length; ++i)
-	{
-		if (G_UNLIKELY(command[i] == '%'))
-		{
-			switch (command[i + 1])
-			{
-			case 'i':
-				replace_with_quoted_string(command, i, "--icon ", garcon_menu_item_get_icon_name(m_item));
-				break;
-
-			case 'c':
-				replace_with_quoted_string(command, i, garcon_menu_item_get_name(m_item));
-				break;
-
-			case 'k':
-				replace_and_free_with_quoted_string(command, i, garcon_menu_item_get_uri(m_item));
-				break;
-
-			case '%':
-				command.erase(i, 1);
-				break;
-
-			case 'f':
-				// unsupported, pass in a single file dropped on launcher
-			case 'F':
-				// unsupported, pass in a list of files dropped on launcher
-			case 'u':
-				// unsupported, pass in a single URL dropped on launcher
-			case 'U':
-				// unsupported, pass in a list of URLs dropped on launcher
-			default:
-				command.erase(i, 2);
-				break;
-			}
-			length = command.length() - 1;
-		}
-	}
+	gchar* uri = garcon_menu_item_get_uri(m_item);
+	gchar* command = xfce_expand_desktop_entry_field_codes(string,
+			nullptr,
+			garcon_menu_item_get_icon_name(m_item),
+			garcon_menu_item_get_name(m_item),
+			uri,
+			garcon_menu_item_requires_terminal(m_item));
+	g_free(uri);
 
 	// Parse and spawn command
-	gchar** argv;
-	bool result = false;
-	GError* error = nullptr;
-	if (g_shell_parse_argv(command.c_str(), nullptr, &argv, &error))
-	{
-		result = xfce_spawn_on_screen(screen,
-				garcon_menu_item_get_path(m_item),
-				argv, nullptr, G_SPAWN_SEARCH_PATH,
-				garcon_menu_item_supports_startup_notification(m_item),
-				gtk_get_current_event_time(),
-				garcon_menu_item_get_icon_name(m_item),
-				&error);
-		g_strfreev(argv);
-	}
+	spawn(screen, command,
+			garcon_menu_item_get_path(m_item),
+			garcon_menu_item_supports_startup_notification(m_item),
+			garcon_menu_item_get_icon_name(m_item));
 
-	if (G_UNLIKELY(!result))
-	{
-		xfce_dialog_show_error(nullptr, error, _("Failed to execute command \"%s\"."), string);
-		g_error_free(error);
-	}
+	g_free(command);
 }
 
 //-----------------------------------------------------------------------------
 
 void Launcher::run(GdkScreen* screen, DesktopAction* action) const
 {
+	// Expand the field codes
 	const gchar* string = action->get_command();
-	if (exo_str_is_empty(string))
+	if (xfce_str_is_empty(string))
 	{
 		return;
 	}
-	std::string command(string);
 
-	// Expand the field codes
-	std::string::size_type length = command.length() - 1;
-	for (std::string::size_type i = 0; i < length; ++i)
-	{
-		if (G_UNLIKELY(command[i] == '%'))
-		{
-			switch (command[i + 1])
-			{
-			case 'i':
-				replace_with_quoted_string(command, i, "--icon ", action->get_icon());
-				break;
-
-			case 'c':
-				replace_with_quoted_string(command, i, action->get_name());
-				break;
-
-			case 'k':
-				replace_and_free_with_quoted_string(command, i, garcon_menu_item_get_uri(m_item));
-				break;
-
-			case '%':
-				command.erase(i, 1);
-				break;
-
-			case 'f':
-				// unsupported, pass in a single file dropped on launcher
-			case 'F':
-				// unsupported, pass in a list of files dropped on launcher
-			case 'u':
-				// unsupported, pass in a single URL dropped on launcher
-			case 'U':
-				// unsupported, pass in a list of URLs dropped on launcher
-			default:
-				command.erase(i, 2);
-				break;
-			}
-			length = command.length() - 1;
-		}
-	}
+	gchar* uri = garcon_menu_item_get_uri(m_item);
+	gchar* command = xfce_expand_desktop_entry_field_codes(string,
+			nullptr,
+			action->get_icon(),
+			action->get_name(),
+			uri,
+			false);
+	g_free(uri);
 
 	// Parse and spawn command
-	gchar** argv;
-	bool result = false;
-	GError* error = nullptr;
-	if (g_shell_parse_argv(command.c_str(), nullptr, &argv, &error))
-	{
-		result = xfce_spawn_on_screen(screen,
-				garcon_menu_item_get_path(m_item),
-				argv, nullptr, G_SPAWN_SEARCH_PATH,
-				garcon_menu_item_supports_startup_notification(m_item),
-				gtk_get_current_event_time(),
-				action->get_icon(),
-				&error);
-		g_strfreev(argv);
-	}
+	spawn(screen, command,
+			garcon_menu_item_get_path(m_item),
+			garcon_menu_item_supports_startup_notification(m_item),
+			action->get_icon());
 
-	if (G_UNLIKELY(!result))
-	{
-		xfce_dialog_show_error(nullptr, error, _("Failed to execute command \"%s\"."), string);
-		g_error_free(error);
-	}
+	g_free(command);
 }
 
 //-----------------------------------------------------------------------------
